/***
 * AOKell E-Store application
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package estore.store;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.api.NoSuchInterfaceException;

import estore.bank.IBank;
import estore.bank.InsufficientBalanceException;
import estore.bank.UnknownAccountException;
import estore.client.IClient;
import estore.provider.IProvider;
import estore.provider.UnknownItemException;

public class StoreImpl
    implements IStore, IFastLane, ILane, IJustHaveALook {

    public final static String IPROVIDER_BINDING = "IProvider";
    public final static String IBANK_BINDING = "IBank";
    private IProvider iProvider;
    private IBank iBank;

    /**
     * Constructs a new StoreImpl
     */
    public StoreImpl() {
        // ...
    }

    // -----------------------------------------------------
    // Implementation of the IJustHaveALook interface
    // -----------------------------------------------------

    /**
     * @param item  a given item
     * @return      the price of a given item
     * @throws UnknownItemException
     */
    public double getPrice( Object item ) throws UnknownItemException {
        return iProvider.getPrice(item);
    }
    
    /**
     * @param item  a given item
     * @param qty   a given quantity
     * @return
     *      true if the given quantity of the given item is available
     *      directly from the store
     *      i.e. without having to re-order it from the provider
     */
    public boolean isAvailable( Object item, int qty )
    throws UnknownItemException {
        
        if ( ! itemsInStock.containsKey(item) )
            throw new UnknownItemException(
                    "Item "+item+
                    " does not correspond to any known reference");
        
        ItemInStock iis = (ItemInStock) itemsInStock.get(item);
        boolean isAvailable = (iis.quantity >= qty);
        
        return isAvailable;
    }

    // -----------------------------------------------------
    // Implementation of the IStore interface
    // -----------------------------------------------------

    // -----------------------------------------------------
    // Implementation of the ILane interface
    // -----------------------------------------------------
    /**
     * Add an item to a cart.
     * If the cart does not exist yet, create a new one.
     * This method is called for each item one wants to add to the cart.
     * 
     * @param cart    a previously created cart or null
     * @param client
     * @param item
     * @param qty
     * @return
     *      Implementation dependant.
     *      Either a new cart at each call or the same cart updated.
     * 
     * @throws UnknownItemException
     * @throws MismatchClientCartException
     *      if the given client does not own the given cart
     */
    public Cart addItemToCart(
            Cart cart,
            IClient client,
            Object item,
            int qty )
    throws UnknownItemException, InvalidCartException {
        
        if ( cart == null ) {
            // If no cart is provided, create a new one
            cart = new Cart(client);
        }
        else {
            if ( client != cart.client )
                throw new InvalidCartException(
                        "Cart "+cart+" does not belong to "+client);
        }
        
        cart.addItem(item,qty);
        
        return cart;
    }

    /**
     * Once all the items have been added to the cart,
     * this method finish make the payment
     *  
     * @param cart
     * @param address
     * @param bankAccount
     * @return  the order
     * 
     * @throws UnknownItemException
     */
    public Order pay( Cart cart, Object address, Object bankAccount )
    throws
    InvalidCartException, UnknownItemException,
    InsufficientBalanceException, UnknownAccountException {
        
        if ( cart == null )
            throw new InvalidCartException("Cart shouldn't be null");
        
        // Create a new order
        Order order = new Order( cart.client, address, bankAccount );
        orders.put( new Integer(order.getKey()), order );
        
        // Order all the items of the cart
        Set entries = cart.getItems().entrySet();
        for (Iterator iter = entries.iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            Object item = entry.getKey();
            int qty = ((Integer) entry.getValue()).intValue();
            
            treatOrder(order,item,qty);            
        }
        double amount = order.computeAmount();
        
        try {
            iBank = (IBank) lookupFc("IBank");
        }
        catch( NoSuchInterfaceException nsie ) {
            throw new RuntimeException(nsie.getMessage());
        }
        
        // Make the payment
        // Throws InsuffisiantBalanceException if the client account is
        // not sufficiently balanced
        iBank.transfert(bankAccount,toString(),amount);
        
        return order;
    }

    // -----------------------------------------------------
    // Implementation of the IFastLane interface
    // -----------------------------------------------------

    /**
     * A map of emitted orders.
     * keys = order keys as Integers
     * values = Order instances
     */
    private Map orders = new HashMap();
    
    /** 
     * A map of items available in the stock of the store.
     * keys = the references of the items as Objects
     * values = ItemInStock instances
     */
    private Map itemsInStock = new HashMap();

    /**
     * Used by a client to order an item.
     * The whole process of ordering is encapsulated by this method.
     * If several items need to be ordered, this method needs to be
     * called several times, but the items will appear in separate orders.
     * 
     * @param client
     * @param item
     * @param qty
     * @param address
     * @param bankAccount
     * @return  the order
     * 
     * @throws UnknownItemException
     * @throws InsufficientBalanceException
     * @throws UnknownAccountException
     */
    public Order oneShotOrder(
            IClient client,
            Object item,
            int qty,
            Object address,
            Object bankAccount
    )
    throws
    UnknownItemException,
    InsufficientBalanceException, UnknownAccountException {
        
        // Create a new order
        Order order = new Order( client, address, bankAccount );
        orders.put( new Integer(order.getKey()), order );
        
        // Treat the item ordered
        treatOrder(order,item,qty);
        double amount = order.computeAmount();
        
        try {
            iBank = (IBank) lookupFc("IBank");
        }
        catch( NoSuchInterfaceException nsie ) {
            throw new RuntimeException(nsie.getMessage());
        }
        
        // Make the payment
        // Throws InsuffisiantBalanceException if the client account is
        // not sufficiently balanced
        iBank.transfert(bankAccount,toString(),amount);
        
        return order;
    }
    
    /**
     * Treat an item ordered by a client and update the corresponding order.
     * 
     * @param order 
     * @param item
     * @param qty
     * @return
     * 
     * @throws UnknownItemException
     * @throws InsufficientBalanceException
     * @throws UnknownAccountException
     */
    private void treatOrder( Order order, Object item, int qty )
    throws UnknownItemException {
        
        // The number of additional item to order
        // in case we need to place an order to the provider
        final int more = 10;

        try {
            iProvider = (IProvider) lookupFc("IProvider");
        }
        catch( NoSuchInterfaceException nsie ) {
            throw new RuntimeException(nsie.getMessage());
        }
        
        // The price of the ordered item
        // Throws UnknownItemException if the item does not exist
        final double price = iProvider.getPrice(item);
        
        final double totalAmount = price*qty;
        
        // The delay (in hours) for delivering the order
        // By default, it takes 2 hours to ship items from the stock
        // This delay increases if an order is to be placed to the provider
        int delay = 2;
        
        // Check whether the item is available in the stock
        // If not, place an order for it to the provider
        ItemInStock iis = (ItemInStock) itemsInStock.get(item);
        if ( iis == null ) {
            int quantity = qty + more;
            delay += iProvider.order(this,item,quantity);
            ItemInStock newItem = new ItemInStock(item,more,price,iProvider);
            itemsInStock.put(item,newItem);
        }
        else {
            // The item is in the stock
            // Check whether there is a sufficient number of them
            // to match the order
            if ( iis.quantity >= qty ) {
                iis.quantity -= qty;
            }
            else {
                // An order to the provider needs to be issued
                int quantity = qty + more;
                delay += iProvider.order(this,item,quantity);
                iis.quantity += more;
            }
        }
        
        // Update the order
        order.addItem(item,qty,price);
        order.setDelay(delay);
    }

    // -----------------------------------------------------
    // Other methods
    // -----------------------------------------------------
    
    public String toString() {
       return "E-Store"; 
    }
}